from PySide2.QtWidgets import QCheckBox
from PySide2.QtCore import Qt

from propsettings.setting import Setting
from propsettings_qt.input_handlers.input_handler import SettingDrawer


class BoolHandler(SettingDrawer):
	"""
	InputHandler que se encarga de las configuraciones de tipo bool a través de un QCheckBox.
	"""

	def __init__(self, setting_owner, setting: Setting):
		super().__init__(setting_owner=setting_owner, setting=setting)
		# validar tipo de dato de la configuración (setting)
		st = self._setting_type
		if st != bool:
			raise TypeError(f'Wrong setting type: {st}. Setting must be of type bool.')

		self.checkbox = _CheckBoxHandler(bool_handler=self)
		self.checkbox.stateChanged.connect(self._on_state_changed)
		self._set_value_to_checkbox()

	def get_widget(self):
		return self.checkbox

	def _set_value_to_checkbox(self):
		value = self._get_value()
		self.checkbox.setChecked(value)

	def _on_state_changed(self, state):
		"""
		Evento ejecutado cuando el valor del checkbox es editado.
		:return:
		"""
		value = state == Qt.CheckState.Checked
		self._set_value(value)


class _CheckBoxHandler(QCheckBox):
	"""
	La única función de esta clase es ser un QCheckBox que tenga como miembro al BoolHandler para que este se quede
	en memoria.
	"""

	def __init__(self, bool_handler, parent=None):
		super().__init__(parent=parent)
		self.bool_handler = bool_handler



from propsettings.setting import Setting
from propsettings_qt.input_handlers.input_handler import SettingDrawer


class ObjectHandler(SettingDrawer):
	"""
	InputHandler que se encarga de las configuraciones que sean de tipo object.
	Estas solo se podrán modificar si tienen miembros de tipo Setting o si son de algún tipo de dato que se puede cargar.
	"""

	def __init__(self, setting_owner, setting: Setting):
		from propsettings_qt.ui_settings_area import SettingsAreaWidget

		super().__init__(setting_owner=setting_owner, setting=setting)
		self.widget = SettingsAreaWidget()
		self.widget.populate_object(setting.fget(self._setting_owner))

	def get_widget(self):
		return self.widget

	def _on_configurable_object_selected(self, obj):
		"""
		Capturar evento de selección de objeto en combobox. Solo se usa cuando el objeto que se está configurando se
		puede cargar desde memoria externa y por tanto se puede seleccionar en un combobox.
		:param obj:
		:return:
		"""
		self._set_value(obj)




from typing import Optional

from PySide2 import QtWidgets

from propsettings_qt import setting_widget_retrieval


class SettingsAreaWidget(QtWidgets.QFrame):
    """
    Widget que muestra las configuraciones de un objeto y permite editarlas.
    """

    def __init__(self, parent=None):
        super(SettingsAreaWidget, self).__init__(parent=parent)
        self._layout = QtWidgets.QVBoxLayout()
        self.setLayout(self._layout)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        self._children_count: int = 0
        self._object_drawer: Optional[QtWidgets.QWidget] = None

    @property
    def children_count(self) -> int:
        return self._children_count

    def populate_object(self, obj):
        """
        Quitar todos los elementos existentes e incializar los widgets de cada Setting de un objeto.
        :param obj:
        :return:
        """
        self.clear()
        self._object_drawer = setting_widget_retrieval.get_object_drawer(obj)
        self._layout.addWidget(self._object_drawer)

        self._children_count = self._get_widget_children_count(self._object_drawer)
        if self._children_count > 0:
            self.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Plain)
        else:
            self.setFrameStyle(QtWidgets.QFrame.NoFrame)

    def _get_widget_children_count(self, widget: QtWidgets.QWidget):
        children_count = 0
        children = widget.children()
        for child in children:
            if isinstance(child, QtWidgets.QWidget):
                children_count += 1
        return children_count

    def clear(self):
        while self._layout.count() > 0:
            item = self._layout.itemAt(0)
            wgt = item.widget()
            if wgt:
                self._layout.removeItem(item)
                wgt.deleteLater()
        self._children_count = 0



from PySide2 import QtWidgets
from propsettings.setting import Setting
from propsettings.setting_types import list_setting_type
from propsettings_qt import setting_widget_retrieval

from propsettings_qt.input_handlers.input_handler import SettingDrawer


class ListHandler(SettingDrawer):

    def __init__(self, list_setting_type: list_setting_type.List, setting_owner, setting: Setting):
        super().__init__(setting_owner=setting_owner, setting=setting)
        # validar tipo de dato de la configuración (setting)
        st = self._setting_type
        if st != list:
            raise TypeError(f'Wrong setting type: {st}. Setting must be of type list.')
        self._setting_type = list_setting_type

    def get_widget(self):
        pass


class SettingsList(QtWidgets.QWidget):

    def __init__(self, list_setting_type: list_setting_type.List, handler: ListHandler, parent=None):
        super(SettingsList, self).__init__(parent=parent)
        self._handler = handler

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

        self._list_widget = QtWidgets.QListWidget()

        self._buttons_widget = QtWidgets.QWidget()
        self._buttons_widget.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        buttons_layout = QtWidgets.QHBoxLayout()
        buttons_layout.setContentsMargins(0, 0, 0, 0)
        self._buttons_widget.setLayout(buttons_layout)

        self._add_button = QtWidgets.QPushButton(text="+")
        self._remove_button = QtWidgets.QPushButton(text="-")
        self._add_button.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        self._remove_button.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)

        buttons_layout.addStretch()
        buttons_layout.addWidget(self._add_button)
        buttons_layout.addWidget(self._remove_button)

    def _add_element(self, element):
        index = 0

        def getter(instance: list):
            return instance[index]

        def setter(instance: list, value):
            instance[index] = value

        stg = Setting(fget=getter, fset=setter, setting_type=setting_type)
        widget = setting_widget_retrieval.get_setting_widget(l, stg)



from PySide2 import QtWidgets, QtCore

from propsettings.setting import Setting
from propsettings.setting_types.selectable_setting_type import Selectable
from propsettings_qt.input_handlers.input_handler import SettingDrawer


class SelectableHandler(SettingDrawer):

    def __init__(self, selectable: Selectable, setting_owner, setting: Setting):
        super(SelectableHandler, self).__init__(setting_owner, setting)
        self._selectable = selectable

        self._selectable_widget = SelectableWidget(self, selectable)
        self._selectable_widget.signal_option_selected.connect(self._on_option_selected)
        self._selectable_widget.populate_options()

    def get_widget(self):
        return self._selectable_widget

    def _on_option_selected(self, option_name: str, option_data: object):
        self._set_value(option_data)
        self._selectable.call_selected_callback(self._setting_owner, option_name, option_data)


class SelectableWidget(QtWidgets.QComboBox):
    signal_option_selected = QtCore.Signal(str, object)

    def __init__(self, handler: SelectableHandler, selectable: Selectable, *args, **kwargs):
        super(SelectableWidget, self).__init__(*args, **kwargs)
        self._handler = handler
        self._selectable = selectable
        self.currentIndexChanged.connect(self._selection_changed)

    def populate_options(self):
        options = self._selectable.options
        for option_name, option_data in options:
            self.addItem(option_name, option_data)

    def _selection_changed(self, index):
        option_name, option_data = self._selectable.options[index]
        self.signal_option_selected.emit(option_name, option_data)



from typing import List

from propsettings import configurable
from propsettings.setting import Setting
from PySide2 import QtWidgets

from propsettings_qt import setting_widget_retrieval
from propsettings_qt.object_drawers.object_drawer import ObjectDrawer


class DefaultObjectDrawer(QtWidgets.QWidget, ObjectDrawer):

    def __init__(self, parent=None):
        super(DefaultObjectDrawer, self).__init__(parent=parent)
        self._layout = QtWidgets.QFormLayout()
        self.setLayout(self._layout)

    def draw_object(self, obj):
        self.clear()
        settings = configurable.get_settings(obj)
        settings = self._sort_settings(settings)
        for stg in settings:
            stg_widget = setting_widget_retrieval.get_setting_widget(obj, stg)
            self._layout.addRow(stg.label, stg_widget)

    def clear(self):
        while self._layout.count() > 0:
            item = self._layout.itemAt(0)
            wgt = item.widget()
            if wgt:
                self._layout.removeItem(item)
                wgt.deleteLater()

    def _sort_settings(self, settings: List[Setting]):
        """
        Ordenar una lista de Settings según los tipos de datos.
        :param settings:
        :return:
        """
        settings_by_sort_order = {}
        for stg in settings:
            settings_by_sort_order.setdefault(stg.sort_order, []).append(stg)

        settings_by_sort_order = sorted(settings_by_sort_order.items(), key=lambda x: x[0])

        sorted_settings = []
        for _, settings in settings_by_sort_order:
            settings.sort(key=lambda stg: stg.label)
            for stg in settings:
                sorted_settings.append(stg)

        return sorted_settings




from abc import ABC, abstractmethod

from propsettings.setting import Setting


class SettingDrawer(ABC):
	"""
	Clase abstracta para implementar widgets que gestionen la entrada para dar valores a atributos de tipo configurable.Setting.
	Cuando los widgets editen el valor de la configuración, deben notificarlo a través del atributo event_value_edited.
	Las clases que hereden de esta no deben ser el widget ellas mismas para evitar multiherencia.
	En cambio, deben devolver el objeto widget a través del método abstracto get_widget.
	"""

	@abstractmethod
	def get_widget(self):
		"""
		Obtener el widget que recibe la entrada del usuario.
		:return:
		"""
		pass

	def __init__(self, setting_owner, setting: Setting):
		"""
		Constructor que recibe el objeto que tiene la configuración y la configuración en sí.
		:param setting_owner: objeto que tiene la configuración. Es necesario para asignarle valor a su configuración
		cuando se edite la configuración en el widget.
		:param setting: configuración (Setting) del objeto
		"""
		self._setting_owner = setting_owner
		self._setting = setting
		self._setting_type = setting.setting_value_type or type(self._get_value())

	def _get_value(self):
		"""
		Obtener el valor del parámetro configurable.
		:return:
		"""
		return self._setting.fget(self._setting_owner)

	def _set_value(self, value):
		"""

		:param value:
		:return:
		"""
		self._setting.fset(self._setting_owner, value)



from PySide2 import QtGui
from PySide2 import QtCore
from PySide2.QtWidgets import QLineEdit

from propsettings.setting import Setting
from propsettings_qt.input_handlers.input_handler import SettingDrawer


class TextHandler(SettingDrawer):
	"""
	InputHandler que se encarga de las configuraciones de tipo int, float y string a través de un QLineEdit.
	"""

	def __init__(self, setting_owner, setting: Setting):
		super().__init__(setting_owner=setting_owner, setting=setting)

		# validar tipo de dato de la configuración (setting)
		st = self._setting_type
		if st != int and st != float and st != str:
			raise TypeError(f'Wrong setting type: {st}. Setting must be of type string (str) or integer (int) or float (float).')

		self.edit = _LineEditHandler(text_handler=self)
		self.edit.editingFinished.connect(self._on_text_edited)
		self._validate_property_type()
		self._set_edit_from_value()

	def get_widget(self):
		return self.edit

	def _validate_property_type(self):
		"""
		Poner validadores al campo de edición para que solo acepte valores del tipo de dato de la propiedad.
		:return:
		"""
		if self._setting_type == int:
			self.edit.setValidator(QtGui.QIntValidator())
		elif self._setting_type == float:
			self.edit.setValidator(QtGui.QDoubleValidator())

	def _get_value_from_edit(self):
		"""
		Obtener valor a partir del string editado en el campo de edición
		:return:
		"""
		value = self.edit.text()
		if self._setting_type == int:
			value = int(value)
		elif self._setting_type == float:
			value = float(value)
		return value

	def _set_edit_from_value(self):
		"""
		Poner el valor de la propiedad en el QLineEdit (campo de edición)
		:return:
		"""
		value = self._get_value()
		self.edit.setText(str(value))

	@QtCore.Slot()
	def _on_text_edited(self):
		"""
		Evento ejecutado cuando el valor del campo de edición es editado.
		:return:
		"""
		value = self._get_value_from_edit()
		self._set_value(value)


class _LineEditHandler(QLineEdit):
	"""
	La única función de esta clase es ser un QLineEdit que tenga como miembro al TextHandler para que este se quede
	en memoria.
	"""

	def __init__(self, text_handler, parent=None):
		super().__init__(parent=parent)
		self.text_handler = text_handler



from PySide2 import QtWidgets, QtCore, QtGui

from propsettings.setting import Setting
from propsettings.setting_types.range_setting_type import Range
from propsettings_qt.input_handlers.input_handler import SettingDrawer
from propsettings_qt.widgets.float_slider import FloatSlider


class RangeHandler(SettingDrawer):

    def __init__(self, value_range: Range, setting_owner, setting: Setting):
        super(RangeHandler, self).__init__(setting_owner, setting)
        self._range = value_range

        self._widget = RangeSlider(self, self._range)
        self._widget.set_value(self._get_value())
        self._widget.signal_value_changed.connect(self._set_value)

    def get_widget(self):
        return self._widget


class RangeSlider(QtWidgets.QSplitter):

    signal_value_changed = QtCore.Signal(float)

    def __init__(self, handler: RangeHandler, value_range: Range, *args, **kwargs):
        super(RangeSlider, self).__init__(*args, **kwargs)
        self._handler = handler
        self._range = value_range

        self._edit = QtWidgets.QLineEdit()

        if self._is_float():
            self._slider = FloatSlider(orientation=QtCore.Qt.Horizontal)
            self._slider.floatValueChanged.connect(self._on_slider_changed)
            self._edit.setValidator(QtGui.QDoubleValidator())
        else:
            self._slider = QtWidgets.QSlider(orientation=QtCore.Qt.Horizontal)
            self._slider.valueChanged.connect(self._on_slider_changed)
            self._edit.setValidator(QtGui.QIntValidator())

        self._slider.setMinimum(self._range.min)
        self._slider.setMaximum(self._range.max)

        self._edit.editingFinished.connect(self._on_edit_changed)

        self.addWidget(self._slider)
        self.addWidget(self._edit)

    def set_value(self, value):
        value = self._clamp_value(value)
        self._set_slider_value(value)
        self._set_edit_value(value)

    def _set_slider_value(self, value):
        self._slider.setValue(value)

    def _set_edit_value(self, value):
        self._edit.setText(f'{value}'[:5])

    def _on_slider_changed(self, value):
        value = self._clamp_value(value)
        self._set_edit_value(value)
        self.signal_value_changed.emit(value)

    def _on_edit_changed(self):
        value = self._get_value_from_edit()
        value = self._clamp_value(value)
        self._set_slider_value(value)
        self.signal_value_changed.emit(value)

    def _get_value_from_edit(self):
        """
        Obtener valor a partir del string editado en el campo de edición
        :return:
        """
        value = self._edit.text()
        if self._is_float():
            return float(value)
        else:
            return int(value)

    def _is_float(self) -> bool:
        min_is_float = isinstance(self._range.min, float)
        max_is_float = isinstance(self._range.max, float)
        return min_is_float or max_is_float

    def _clamp_value(self, value):
        value = min(self._range.max, max(self._range.min, value))
        return value



from typing import Optional, Dict, Type, Any

from PySide2 import QtWidgets

from propsettings.setting import Setting
from propsettings.setting_type import SettingType
from propsettings.setting_types.range_setting_type import Range
from propsettings.setting_types.selectable_setting_type import Selectable
from propsettings_qt.input_handlers.bool_handler import BoolHandler
from propsettings_qt.input_handlers.input_handler import SettingDrawer
from propsettings_qt.input_handlers.object_handler import ObjectHandler
from propsettings_qt.input_handlers.range_handler import RangeHandler
from propsettings_qt.input_handlers.selectable_handler import SelectableHandler
from propsettings_qt.input_handlers.text_handler import TextHandler
from propsettings_qt.object_drawers.default_drawer import DefaultObjectDrawer
from propsettings_qt.object_drawers.object_drawer import ObjectDrawer


object_drawers: Dict[type, Type[ObjectDrawer]] = {
}

setting_type_drawers: Dict[Type[SettingType], Type[SettingDrawer]] = {
	Range: RangeHandler,
	Selectable: SelectableHandler,
}

setting_value_type_handlers: Dict[type, Type[SettingDrawer]] = {
	bool: BoolHandler,
	int: TextHandler,
	float: TextHandler,
	str: TextHandler,
}


def register_object_drawer(object_type: type, object_drawer_type: Type[ObjectDrawer]):
	if object_type not in object_drawers:
		if issubclass(object_drawer_type, QtWidgets.QWidget):
			object_drawers[object_type] = object_drawer_type


def register_setting_type_drawer(setting_type: Type[SettingType], handler_type: Type[SettingDrawer]):
	if setting_type not in setting_type_drawers:
		setting_type_drawers[setting_type] = handler_type


def get_object_drawer(obj) -> QtWidgets.QWidget:
	obj_type = type(obj)
	obj_drawer = None
	if obj_type in object_drawers:
		drawer_type = object_drawers[obj_type]
		if issubclass(drawer_type, QtWidgets.QWidget):
			widget = drawer_type()
			obj_drawer = widget
	if obj_drawer is None:
		obj_drawer = DefaultObjectDrawer()
	obj_drawer.draw_object(obj)
	return obj_drawer


def get_setting_widget(obj, setting: Setting) -> QtWidgets.QWidget:
	"""
	Obtener el QWidget apropiado para editar una setting.
	:param obj:
	:param setting:
	:return:
	"""
	widget = _get_widget_by_setting_type(obj, setting)

	# if it hasn't setting type, get widget by value type
	if widget is None:
		widget = _get_widget_by_value_type(obj, setting)

	# decorate widget with setting's decorators
	# TODO

	return widget


def _get_widget_by_setting_type(obj, setting: Setting) -> Optional[QtWidgets.QWidget]:
	setting_type = setting.setting_type
	if setting_type is not None:
		setting_type_type = type(setting_type)
		if setting_type_type in setting_type_drawers:
			handler_class = setting_type_drawers[setting_type_type]
			handler = handler_class(setting_type, obj, setting)
			widget = handler.get_widget()
			return widget
	return None


def _get_widget_by_value_type(obj, setting: Setting) -> QtWidgets.QWidget:
	setting_value_type = setting.setting_value_type or type(setting.fget(obj))
	if setting_value_type in setting_value_type_handlers:
		handler_class: type = setting_value_type_handlers[setting_value_type]
		handler: SettingDrawer = handler_class(obj, setting)
		widget = handler.get_widget()
	else:
		widget = ObjectHandler(obj, setting).get_widget()
	return widget
